import csv, re, time


def multireplace(string, replacements):
    """
    Given a string and a replacement map, it returns the replaced string.
    :param str string: string to execute replacements on
    :param dict replacements: replacement dictionary {value to find: value to replace}
    :rtype: str
    """
    # Place longer ones first to keep shorter substrings from matching where the longer ones should take place
    # For instance given the replacements {'ab': 'AB', 'abc': 'ABC'} against the string 'hey abc', it should produce
    # 'hey ABC' and not 'hey ABc'
    substrs = sorted(replacements, key=len, reverse=True)

    # Create a big OR regex that matches any of the substrings to replace
    regexp = re.compile('|'.join(map(re.escape, substrs)))

    # For each match, look up the new string in the replacements
    return regexp.sub(lambda match: replacements[match.group(0)], string)


# Ton data set from http://ltr.uio.no/ton/ton.data.101.tgz
talk_of_norway = []
file_counter = 1

'''
Typical Norwegian Parliament abbreviations or otherwise typical abbreviations that causes a bad split by the tokenizer.
'''
replacement_map = {
    'pst.': 'prosent',
    'kr.': 'kroner',
    'mill.': 'mill',
    'st. meld.': 'stortingsmelding',
    'St. meld.': 'Stortingsmelding',
    'st.meld.': 'stortingsmelding',
    'St.meld.': 'Stortingsmelding',
    'ot. meld.': 'odelstingsmelding',
    'Ot. meld.': 'Odelstingsmelding',
    'ot.meld.': 'odelstingsmelding',
    'Ot.meld.': 'Odelstingsmelding',
    'St.prp.': 'Stortingsproposisjon',
    'st.prp.': 'stortingsproposisjon',
    'St. prp.': 'Stortingsproposisjon',
    'st. prp.': 'stortingsproposisjon',
    'Ot.prp.': 'Odelstingsproposisjon',
    'ot.prp.': 'odelstingsproposisjon',
    'Ot. prp.': 'Odelstingsproposisjon',
    'ot. prp.': 'odelstingsproposisjon',
    'nr.': 'nummer',
    'bl.a.': 'blant annet',
    'bl. a.': 'blant annet',
    'jf.': 'jf ',
    'm.a.o.': 'med andre ord',
    'm.fl.': 'med flere',
    'm. fl.': 'med flere',
    'mfl.': 'med flere',
    'kgl.': 'kongelig',
    'f. eks.': 'for eksempel',
    'f.eks.': 'for eksempel',
    'Innst.': 'Innstilling',
    'p.g.a.': 'på grunn av',
    'mht.': 'med hensyn til',
    'osb.': 'og så bortetter',
    'mv.': 'med videre'
}

removable_all_words = ['følgende', 'representant', 'innkall']  # DO NOT add more words here. Because ALL phrases in this list must be present in a sentence for it to be removed.

removable_all_words2 = ['møte', 'innkall']


# Specific speeches to remove based on manual filtering
speeches_to_remove = ['tale014459', 'tale013935', 'tale013936', 'tale014457', 'tale014464', 'tale014491', 'tale014185', 'tale014508', 'tale002250', 'tale002745']  # Talks alot about prices, limits, tobacco, alcohol, or sums stuff

party_list = []

start = time.clock()
current_party = ''

party_dict = {"A": [], "FrP": [], "H": [], "Kp": [], "KrF": [], "MDG": [], "NA": [], "Sp": [], "SV": [], "TF": [], "Uavhengig": [], "V": []}


with open('ton.csv', 'r', encoding='utf-8') as ton_file:
    reader = csv.reader(ton_file)
    next(reader)  # skip header
    sortedlist = sorted(reader, key=lambda row: row[11], reverse=True)
    for row in sortedlist:
        id = row[0]
        rep_name = row[5]
        party = row[11]
        date = row[81]
        language = row[76]
        if language == 'nno':
            continue
        if rep_name.lower() == 'presidenten' and len(party) == 0:
            continue
        party = row[11] if len(row[11]) != 0 else 'NA'
        # if party == 'NA' and (rep_name == 'NA' or len(rep_name) == 0):
        #    print('party = 0 and rep_name na or 0: {} and {}'.format(party, rep_name))
        if all(w in row[83].lower() for w in removable_all_words):
            continue  # if we find all values in removable_all_words, jump to next iteration
        if all(w in row[83].lower() for w in removable_all_words2):
            continue  # if we find all values in removable_all_words, jump to next iteration
        no_no_break_space_text = row[83].replace(u'\xa0', ' ')  # Unicode NO-BREAK SPACE. Just replace with space
        text = multireplace(no_no_break_space_text, replacement_map)
        party_dict[party].append([id, party, rep_name, date, text])

print('Finished filtering data, writing to files...')

for key, value in party_dict.items():
    with open('party_filtered/ton_filtered_{}.csv'.format(key), 'a', encoding='utf-8') as ton_out_file:
        writer = csv.writer(ton_out_file, delimiter=',', lineterminator='\n')
        for ton in value:
            if ton[0] not in speeches_to_remove:
                writer.writerow([ton[0], ton[1], ton[2], ton[3], ton[4]])
            else:
                print('Found {} at pos {}. Skipping..'.format(ton[0], speeches_to_remove.index(ton[0])))
end = time.clock()

print('Done. Time used: {}'.format(end-start))
