var express = require('express');
var app = express();
var cors = require('cors');
var bodyParser = require('body-parser');
var bcrypt = require('bcrypt');
var jwt = require('jsonwebtoken');
var sqlite3 = require('sqlite3').verbose();
var winston = require('winston');
var port;
var URLS = require('../src/urls');

var answer_count_limit = 3;
require('dotenv').load(); // load .env variables

if (process.env.NODE_ENV==='development') {
    port = 8000;
    app.use(cors({credentials: true, origin: true}))
}else {
    port = 3000;
    var serveStatic = require('serve-static');
    // Make all static urls accessible
    app.use('/', serveStatic('../build'));
    app.use('/callback', serveStatic('../build'));
    app.use('/home', serveStatic('../build'));
    app.use('/classification', serveStatic('../build'));
    app.use('/classification-training', serveStatic('../build'));
    app.use('/done', serveStatic('../build'));
    app.use('/results', serveStatic('../build'));
    app.use('/thanks', serveStatic('../build'));
    app.use('/login', serveStatic('../build'));
    app.use('/create-user', serveStatic('../build'));

    // Terms of service and privacy policy
    app.use('/terms-and-conditions', serveStatic('../tos.html'));
    app.use('/privacy', serveStatic('../privacy.html'));
}

console.log('URLS: ', URLS);


/*
 Base logger for backend errors
 */
let logger = winston.createLogger({
    format: winston.format.json(),
    transports: [
        //
        // - Write to all logs with level `info` and below to `combined.log`
        // - Write all logs error (and below) to `error.log`.
        //
        new winston.transports.File({ filename: './logs/winston_api_main_errors.log', level: 'error'}),
        new winston.transports.File({ filename: './logs/winston_api_other_errors.log', level: 'info' })
    ]
});

/*
 Base logger for API requests
 */
let logger2 = winston.createLogger({
    format: winston.format.json(),
    transports: [
        //
        // - Write to all logs with level `info` and below to `combined.log`
        // - Write all logs error (and below) to `error.log`.
        //
        new winston.transports.File({ filename: './logs/winston_req.log'}),
    ]
});

var frontend_logger = winston.createLogger({
    format: winston.format.json(),
    transports: [
        //
        // - Write to all logs with level `info` and below to `combined.log`
        // - Write all logs error (and below) to `error.log`.
        //
        new winston.transports.File({ filename: './logs/winston_frontend_error.log', level: 'error' }),
        new winston.transports.File({ filename: './logs/winston_fronted_combined.log' })
    ]
});

var request_status = function (req, res, next) {
    var headers = req.headers;
    if (headers['authorization']){
        getInternalUsername(headers['authorization'], function (username) {
            logger2.log({
                timestamp: Date.now(),
                level: 'info',
                username: username,
                origin: headers.origin || 'n/a',
                url: req.originalUrl,
                params: req.params,
                query: req.query,
                data: req.body
            });
            next();
        })
    }else {
        logger2.log({
            timestamp: Date.now(),
            level: 'info',
            username: 'n/a',
            origin: headers.origin || 'n/a',
            url: req.originalUrl,
            params: req.params,
            query: req.query,
            data: req.body
        });
        next();
    }

};

app.use(request_status);

var distributed_usernames = {};
var valid_age_groups = {'age18':true, 'age19-29':true, 'age30-39':true, 'age40-49':true, 'age50-59':true, 'age60':true};
var valid_genders = {'mann':true, 'kvinne':true, 'annet':true};
var valid_eduaction = {'humaniora':true, 'realfag':true, 'rettsvitenskap':true, 'medisin':true, 'samfunnsvitenskap':true, 'kunst':true, 'idrett':true, 'yrkesfag':true, 'annet':true};
var distributed_distcodes = {};

/*var whitelist = [URLS.app_url, URLS.ssl_app_url];
var corsOptions = {
    credentials: true,
    origin: function (origin, callback) {
        if (whitelist.indexOf(origin) !== -1) {
            callback(null, true);
        } else {
            callback({status: 403}, false);
        }
    }
};*/

// configure body parser
app.use(bodyParser.urlencoded({ extended: true }));



function signJWT(username, expires_in) {
    var payload = {
        username: username
    };
    var options = {
        audience: URLS.callable_api_url,
        issuer: "https://claimcollector.idi.ntnu.no/",
        algorithms: ['RS256']
    };
    var expires_at = ((Math.floor((new Date()).getTime()/1000))+expires_in)*1000;
    var token = jwt.sign(payload, process.env.JWTSECRET, {
        expiresIn: expires_in // expires in 24 hours
    });
    return [token, expires_at];
}


var jwtCheck = function (req, res, next) {
    var authorization = req.headers['authorization'];
    if (authorization) {
        var token = parseToken(authorization);
        if (token) {
            jwt.verify(token, process.env.JWTSECRET, function (err, decoded) {
                if (err) {
                    logger.log({
                        timestamp: new Date().getTime(),
                        level: 'error',
                        message: 'Error in jwtCheck 1',
                        exception: err,
                        token: token,
                        values: {
                            origin: req.headers.origin,
                            originalUrl: req.originalUrl

                        }
                    });
                    next();
                    //res.status(401).json({message: 'Unauthorized'});
                }else {
                    next();
                }
            })
        }else {
            logger.log({
                timestamp: new Date().getTime(),
                level: 'info',
                message: 'Error in jwtCheck 2',
                exception: 'No token',
                values: {
                    origin: req.headers.origin,
                    originalUrl: req.originalUrl

                }
            });
            res.status(401).json({message: 'Invalid token'});
        }
    }else {
        logger.log({
            timestamp: new Date().getTime(),
            level: 'info',
            message: 'Error in jwtCheck 3',
            exception: 'No authorization header',
            values: {
                origin: req.headers.origin,
                originalUrl: req.originalUrl

            }
        });
        res.status(401).json({message: 'Invalid token'});
    }
};


function parseToken(authorization) {
    if (authorization) {
        var parts = authorization.split(' ');
        if (parts.length === 2) {
            var scheme = parts[0];
            var credentials = parts[1];
            if (/^Bearer$/i.test(scheme)) {
                return credentials.slice(1,-1);
            }
            else{
                logger.log({
                    timestamp: new Date().getTime(),
                    level: 'info',
                    message: 'Error in parseToken',
                    authorization: authorization,
                    scheme: scheme,
                    credentials: credentials
                });
                return null;
            }
        }
    }else {
        return null;
    }
}


function getInternalUsername(authorization, callback) {
    var token = parseToken(authorization);
    if (token) {
        jwt.verify(token, process.env.JWTSECRET, function (err, decoded) {
            if (err) {
                logger.log({
                    timestamp: new Date().getTime(),
                    level: 'info',
                    message: 'Error in getInternalUsername 1',
                    exception: err
                });
                return callback(null);
            }else {
                //console.log('TOKEN IS VALID: ', decoded.username);
                //req.decoded = decoded; // can use decoded to get username
                return callback(decoded.username);
            }
        })
    }else {
        logger.log({
            timestamp: new Date().getTime(),
            level: 'info',
            message: 'Error in getInternalUsername 2',
            exception: 'No token'
        });
        return callback(null);
    }
}

/*
 Fisher–Yates shuffle algorithm that works in linux
 */
function shuffle(array) {
    var currentIndex = array.length, temporaryValue, randomIndex;

    // While there remain elements to shuffle...
    while (0 !== currentIndex) {

        // Pick a remaining element...
        randomIndex = Math.floor(Math.random() * currentIndex);
        currentIndex -= 1;

        // And swap it with the current element.
        temporaryValue = array[currentIndex];
        array[currentIndex] = array[randomIndex];
        array[randomIndex] = temporaryValue;
    }

    return array;
}


function shuffleQuestionsWithBaseline(questions, baseline, num_baseline_questions) {
    //const shuffled_questions = questions.sort(() => .5 - Math.random());// shuffle
    const shuffled_questions = shuffle(questions);// shuffle

    //const shuffled_baseline = baseline.sort(() => .5 - Math.random());// shuffle
    const shuffled_baseline = shuffle(baseline); // shuffle

    var selected_baseline = shuffled_baseline.slice(0, num_baseline_questions); //get sub-array of first n elements AFTER shuffle

    return {'questions': shuffled_questions, 'baseline': selected_baseline};
}




/*
 Calculate precision for users based on baseline questions
 */
function calculateUserScore(username) {
    var db = new sqlite3.Database('ccdb01.db');

    var final_score = 0;

    // Prepared statements for queries
    var stmt = db.prepare("SELECT a.id as answer_id, answer, control_id, username, bq.claim, bq.correct_answer FROM Answers as a INNER JOIN ControlClaims as bq ON bq.id = a.control_id WHERE username = ? AND control_id IS NOT NULL");
    var stmt2 = db.prepare("SELECT username, score FROM Users WHERE username = ?");

    // Serialize to guarantee query execution before db is closed
    db.serialize(function () {
        /*
         First query: Join Answers table and BaselineQuestions table
         */
        db.run("BEGIN TRANSACTION");
        stmt.all(username, function (err, row) {
            if (err === null) {
                for(var r of row){
                    if (r.correct_answer == 0 && r.answer == 0) {
                        //console.log(`Row is ${JSON.stringify(r, null, 4)}. \nGiving +1\n\n`);
                        final_score+=1
                    }
                    else if (r.correct_answer == 1 && r.answer == 1) {
                        //console.log(`Row is ${JSON.stringify(r, null, 4)}. \nGiving +1\n\n`);
                        final_score+=1
                    }
                    else if (r.correct_answer == 1 && r.answer == 2) {
                        //console.log(`Row is ${JSON.stringify(r, null, 4)}. \nGiving +0.6\n\n`);
                        final_score+=0.6
                    }
                    else if (r.correct_answer == 2 && r.answer == 1) {
                        //console.log(`Row is ${JSON.stringify(r, null, 4)}. \nGiving +0.4\n\n`);
                        final_score+=0.4
                    }
                    else if (r.correct_answer == 2 && r.answer == 2) {
                        //console.log(`Row is ${JSON.stringify(r, null, 4)}. \nGiving +1\n\n`);
                        final_score+=1
                    }
                }

                //console.log('score1: ', final_score);
                final_score = parseFloat((final_score/row.length).toFixed(2));
                //console.log(`final/num: ${final_score} num:(${row.length})`);
                var user_exists = null;

                /*
                 Second query: Get all userscores for this user (username = this)
                 */
                stmt2.all(username, function (err, row) {
                    if(err === null) {
                        //console.log('user score row: ', row);
                        user_exists = row.length !== 0;
                        //console.log('User exists in stmt2: ', row.length != 0)
                        /*
                         Third query: If user exists: update UserScores table; else: add to UserScores table
                         */
                        //console.log('User exists before check: ', user_exists);
                        if (user_exists) {
                            var stmt3 = db.prepare("UPDATE Users SET score = ? WHERE username = ?");
                            stmt3.run(final_score, username);
                            stmt3.finalize();
                            //console.log(`Done UPDATING UserScore with score = ${final_score} for user ${username}`)
                        }
                        else {
                            logger.log({
                                timestamp: new Date().getTime(),
                                level: 'error',
                                message: 'Error occurred in calculateUserScore (stmt3)',
                                exception: err
                            });
                        }
                    }
                    else {
                        logger.log({
                            timestamp: new Date().getTime(),
                            level: 'error',
                            message: 'Error occurred in calculateUserScore (stmt2)',
                            exception: err
                        });
                    }
                });
                stmt2.finalize();
            }
            else {
                stmt2.finalize();
                logger.log({
                    timestamp: new Date().getTime(),
                    level: 'error',
                    message: 'Error occurred in calculateUserScore (stmt)',
                    exception: err
                });
            }
        });
        stmt.finalize();
        db.run("COMMIT TRANSACTION");
    });

    db.close();
}

app.get('/api/generate-username', function (req, res) {
    if (req.headers['button-event'] !== 'dkjfgh908u3jrnfp89dyduh34gbhjait78'){
        logger.log({
            timestamp: new Date().getTime(),
            level: 'info',
            message: 'Tried generate-username with invalid button-event'
        });
        res.status(403).json({message: 'Forbidden'})
    }
    else {
        var db = new sqlite3.Database('ccdb01.db');

        var animals = ['Knoppsvane', 'Dvergsvane', 'Sangsvane', 'Sædgås', 'Kortnebbgås', 'Tundragås', 'Dverggås', 'Grågås', 'Kanadagås', 'Hvitkinngås', 'Ringgås', 'Rødhalsgås', 'Niland', 'Rustand', 'Gravand', 'Mandarinand', 'Brunnakke', 'Snadderand', 'Krikkand', 'Amerikakrikkand', 'Stokkand', 'Rødfotand', 'Stjertand', 'Knekkand', 'Blåvingeand', 'Skjeand', 'Rødhodeand', 'Taffeland', 'Ringand', 'Hvitøyeand', 'Toppand', 'Bergand', 'Purpurhodeand', 'Ærfugl', 'Praktærfugl', 'Brilleærfugl', 'Stellerand', 'Harlekinand', 'Havelle', 'Svartand', 'Brilleand', 'Sjøorre', 'Knoppsjøorre', 'Islandsand', 'Kvinand', 'Lappfiskand', 'Siland', 'Laksand', 'Stivhaleand', 'Jerpe', 'Lirype', 'Fjellrype', 'Orrfugl', 'Storfugl', 'Rapphøne', 'Vaktel', 'Smålom', 'Storlom', 'Islom', 'Gulnebblom', 'Ringnebbdykker', 'Dvergdykker', 'Toppdykker', 'Gråstrupedykker', 'Horndykker', 'Svarthalsdykker', 'Svartbrynalbatross', 'Gulnesealbatross', 'Havhest', 'Flekkpetrell', 'Silkepetrell', 'Gulnebblire', 'Storlire', 'Grålire', 'Havlire', 'Havsvale', 'Stormsvale', 'Japanstormsvale', 'Havsule', 'Storskarv', 'Toppskarv', 'Rørdrum', 'Amerikarørdrum', 'Dvergrørdrum', 'Natthegre', 'Kuhegre', 'Silkehegre', 'Egretthegre', 'Gråhegre', 'Purpurhegre', 'Svartstork', 'Stork', 'Bronseibis', 'Skjestork', 'Vepsevåk', 'Svartglente', 'Glente', 'Båndhavørn', 'Havørn', 'Åtselgribb', 'Slangeørn', 'Sivhauk', 'Myrhauk', 'Steppehauk', 'Enghauk', 'Hønsehauk', 'Spurvehauk', 'Musvåk', 'Ørnvåk', 'Fjellvåk', 'Storskrikørn', 'Steppeørn', 'Keiserørn', 'Kongeørn', 'Fiskeørn', 'Rødfalk', 'Tårnfalk', 'Aftenfalk', 'Dvergfalk', 'Lerkefalk', 'Jaktfalk', 'Vandrefalk', 'Vannrikse', 'Myrrikse', 'Sumprikse', 'Åkerrikse', 'Sivhøne', 'Amerikapurpurhøne', 'Sultanhøne', 'Sothøne', 'Trane', 'Kanadatrane', 'Jomfrutrane', 'Dvergtrappe', 'Stortrappe', 'Tjeld', 'Stylteløper', 'Avosett', 'Triel', 'Ørkenløper', 'Brakksvale', 'Orientbrakksvale', 'Steppebrakksvale', 'Dverglo', 'Sandlo', 'Amerikasandlo', 'Tobeltelo', 'Hvitbrystlo', 'Mongollo', 'Ørkenlo', 'Rødbrystlo', 'Boltit', 'Sibirlo', 'Kanadalo', 'Heilo', 'Tundralo', 'Sumpvipe', 'Vipe', 'Sibirsnipe', 'Polarsnipe', 'Sandløper', 'Sandsnipe', 'Beringsnipe', 'Rødstrupesnipe', 'Dvergsnipe', 'Bonapartesnipe', 'Gulbrystsnipe', 'Alaskasnipe', 'Spisshalesnipe', 'Tundrasnipe', 'Fjæreplytt', 'Myrsnipe', 'Styltesnipe', 'Fjellmyrløper', 'Rustsnipe', 'Brushane', 'Kvartbekkasin', 'Enkeltbekkasin', 'Dobbeltbekkasin', 'Rugde', 'Svarthalespove', 'Svartvingespove', 'Lappspove', 'Dvergspove', 'Småspove', 'Storspove', 'Præriesnipe', 'Tereksnipe', 'Strandsnipe', 'Flekksnipe', 'Skogsnipe', 'Eremittsnipe', 'Sotsnipe', 'Plystresnipe', 'Gluttsnipe', 'Gulbeinsnipe', 'Willetsnipe', 'Damsnipe', 'Grønnstilk', 'Rødstilk', 'Steinvender', 'Hvithalesvømmesnipe', 'Svømmesnipe', 'Polarsvømmesnipe', 'Polarjo', 'Tyvjo', 'Fjelljo', 'Storjo', 'Svartehavsmåke', 'Lattermåke', 'Franklinmåke', 'Kanadahettemåke', 'Hettemåke', 'Ringnebbmåke', 'Fiskemåke', 'Sildemåke', 'Gråmåke', 'Kaspimåke', 'Gulbeinmåke', 'Gråvingemåke', 'Eskimomåke', 'Grønlandsmåke', 'Polarmåke', 'Svartbak', 'Rosenmåke', 'Krykkje', 'Ismåke', 'Sabinemåke', 'Dvergmåke', 'Sandterne', 'Rovterne', 'Kongeterne', 'Splitterne', 'Rosenterne', 'Makrellterne', 'Rødnebbterne', 'Tøyleterne', 'Dvergterne', 'Hvitkinnsvartterne', 'Svartterne', 'Hvitvingesvartterne', 'Lomvi', 'Polarlomvi', 'Alke', 'Teist', 'Alkekonge', 'Lunde', 'Steppehøne', 'Bydue', 'Skogdue', 'Ringdue', 'Tyrkerdue', 'Turteldue', 'Mongolturteldue', 'Skjæregjøk', 'Gjøk', 'Gulnebbgjøk', 'Tårnugle', 'Dverghornugle', 'Hubro', 'Snøugle', 'Haukugle', 'Spurveugle', 'Kattugle', 'Slagugle', 'Lappugle', 'Hornugle', 'Jordugle', 'Perleugle', 'Nattravn', 'Skorsteinsseiler', 'Pigghaleseiler', 'Tårnseiler', 'Gråseiler', 'Alpeseiler', 'Kafferseiler', 'Isfugl', 'Bieter', 'Blåkinnbieter', 'Blåråke', 'Hærfugl', 'Vendehals', 'Gråspett', 'Grønnspett', 'Svartspett', 'Flaggspett', 'Hvitryggspett', 'Dvergspett', 'Tretåspett', 'Hvitvingelerke', 'Dverglerke', 'Flekkdverglerke', 'Topplerke', 'Trelerke', 'Sanglerke', 'Fjellerke', 'Sandsvale', 'Klippesvale', 'Låvesvale', 'Taksvale', 'Amursvale', 'Tartarpiplerke', 'Mongolpiplerke', 'Markpiplerke', 'Sibirpiplerke', 'Trepiplerke', 'Tundrapiplerke', 'Heipiplerke', 'Myrpiplerke', 'Vannpiplerke', 'Skjærpiplerke', 'Gulerle', 'Sitronerle', 'Vintererle', 'Linerle', 'Sidensvans', 'Fossekall', 'Gjerdesmett', 'Jernspurv', 'Sibirjernspurv', 'Alpejernspurv', 'Hekkskvett', 'Rødstrupe', 'Nattergal', 'Sørnattergal', 'Rubinstrupe', 'Blåstrupe', 'Blåstjert', 'Hvitstrupenattergal', 'Svartrødstjert', 'Rødstjert', 'Buskskvett', 'Asiasvartstrupe', 'Svartstrupe', 'Isabellasteinskvett', 'Steinskvett', 'Svartstrupesteinskvett', 'Middelhavssteinskvett', 'Ørkensteinskvett', 'Steintrost', 'Gulltrost', 'Sibirtrost', 'Ringtrost', 'Svarttrost', 'Gråstrupetrost', 'Bruntrost', 'Svartflekktrost', 'Svartstrupetrost', 'Gråtrost', 'Måltrost', 'Rødvingetrost', 'Duetrost', 'Starrsanger', 'Stripesanger', 'Gresshoppesanger', 'Elvesanger', 'Sumpsanger', 'Tykknebbsanger', 'Tartarsanger', 'Ramasanger', 'Bleksanger', 'Gulsanger', 'Spottesanger', 'Vannsanger', 'Sivsanger', 'Åkersanger', 'Busksanger', 'Myrsanger', 'Rørsanger', 'Trostesanger', 'Rødstrupesanger', 'Svarthodesanger', 'Mestersanger', 'Hauksanger', 'Møller', 'Tornsanger', 'Hagesanger', 'Munk', 'Østkronsanger', 'Østsanger', 'Lappsanger', 'Fuglekongesanger', 'Gulbrynsanger', 'Blekbrynsanger', 'Viersanger', 'Brunsanger', 'Eikesanger', 'Furusanger', 'Bøksanger', 'Gransanger', 'Løvsanger', 'Fuglekonge', 'Rødtoppfuglekonge', 'Gråfluesnapper', 'Dvergfluesnapper', 'Halsbåndfluesnapper', 'Svarthvit fluesnapper', 'Skjeggmeis', 'Stjertmeis', 'Løvmeis', 'Granmeis', 'Lappmeis', 'Toppmeis', 'Svartmeis', 'Blåmeis', 'Kjøttmeis', 'Spettmeis', 'Trekryper', 'Pungmeis', 'Pirol', 'Brunvarsler', 'Rødhalevarsler', 'Tornskate', 'Rosenvarsler', 'Varsler', 'Krattvarsler', 'Rødhodevarsler', 'Hvitpannevarsler', 'Nøtteskrike', 'Lavskrike', 'Skjære', 'Nøttekråke', 'Kaie', 'Kornkråke', 'Svartkråke', 'Kråke', 'Ravn', 'Stær', 'Rosenstær', 'Gråspurv', 'Middelhavsspurv', 'Pilfink', 'Bokfink', 'Bjørkefink', 'Gulirisk', 'Grønnfink', 'Grønnsisik', 'Tornirisk', 'Bergirisk', 'Brunsisik', 'Gråsisik', 'Polarsisik', 'Båndkorsnebb', 'Grankorsnebb', 'Furukorsnebb', 'Rosenfink', 'Konglebit', 'Dompap', 'Kjernebiter', 'Gulbrynkjernebiter', 'Myrteparula', 'Svarthetteparula', 'Ovnparula', 'Sangspurv', 'Hvitkronespurv', 'Hvitstrupespurv', 'Vinterjunko', 'Lappspurv', 'Snøspurv', 'Gråhodespurv', 'Hvithodespurv', 'Gulspurv', 'Rødbrystspurv', 'Hortulan', 'Vierspurv', 'Dvergspurv', 'Kastanjespurv', 'Sibirspurv', 'Sivspurv', 'Svarthodespurv', 'Kornspurv', 'Sisselspurv', 'Bobolink', 'Lundtrupial', 'Piggsvin', 'Spissmus', 'Knøttspissmus', 'Dvergspissmus', 'Lappspissmus', 'Krattspissmus', 'Taigaspissmus', 'Vannspissmus', 'Hagespissmus', 'Vannflaggermus', 'Børsteflaggermus', 'Skjeggflaggermus', 'Skogflaggermus', 'Skimmelflaggermus', 'Nordflaggermus', 'Sørflaggermus', 'Storflaggermus', 'Dvergflaggermus', 'Tusseflaggermus', 'Trollflaggermus', 'Brunlangøre', 'Bredøre', 'Villsvin', 'Moskusfe', 'Dåhjort', 'Hjort', 'Elg', 'Rådyr', 'Rein', 'Nordspisshval', 'Nebbhval', 'Gåsenebbhval', 'Hvithval', 'Narhval', 'Spermhval', 'Nise', 'Spekkhogger', 'Delfin', 'Stripedelfin', 'Gulflankedelfin', 'Tumler', 'Kvitskjeving', 'Kvitnos', 'Rissodelfin', 'Grindhval', 'Retthvaler', 'Nordkaper', 'Grønlandshval', 'Finnhvaler', 'Vågehval', 'Seihval', 'Finnhval', 'Blåhval', 'Knølhval', 'Gaupe', 'Ulv', 'Mårhund', 'Rødrev', 'Fjellrev', 'Røyskatt', 'Snømus', 'Mink', 'Ilder', 'Mår', 'Jerv', 'Grevling', 'Oter', 'Bjørn', 'Brunbjørn', 'Isbjørn', 'Hvalross', 'Steinkobbe', 'Sel', 'Ringsel', 'Grønlandssel', 'Storkobbe', 'Havert', 'Klappmyss', 'Kanin', 'Sørhare', 'Hare', 'Ekorn', 'Bever', 'Bjørkemus', 'Dvergmus', 'Storskogmus', 'Småskogmus', 'Husmus', 'Brunrotte', 'Svartrotte', 'Hamster', 'Skoglemen', 'Lemen', 'Rødmus', 'Klatremus', 'Gråsidemus', 'Vånd', 'Bisam', 'Markmus', 'Østmarkmus', 'Fjellmarkmus'];
        //['Katt', 'Hund', 'Rotte', 'Kanin', 'Sel', 'Kenguru', 'Ape', 'Bever', 'Hest'];

        var adjectives = ['Advarende', 'Aktiv', 'Aktiviserende', 'Anerkjent', 'Arbeidsom', 'Arbeidende', 'Avgjørende', 'Avhengig', 'Avslørende', 'Avslørt', 'Begynnende', 'Behagelig', 'Behersket', 'Bekymringsfull', 'Beskrivende', 'Beskyttende', 'Bestemt', 'Betydningsfull', 'Betydelig', 'Beveget', 'Bevegelig', 'Bevegende', 'Blomstrende', 'Blomstret', 'Bråkete', 'Bedende', 'Dekket', 'Dikterisk', 'Diktende', 'Drømmende', 'Drømt', 'Endret', 'Engasjerende', 'Engasjert', 'Fantasifull', 'Fantastisk', 'Fascinert', 'Fascinerende', 'Filosoferende', 'Filosofisk', 'Flygende', 'Flytende', 'Fokusert', 'Forbedret', 'Forberedende', 'Forberedt', 'Forenklet', 'Forenklende', 'Forklarende', 'Forklarlig', 'Forlenget', 'Forlengende', 'Forståelig', 'Forståelsesfull', 'Forsvarende', 'Forsøkende', 'Fortellende', 'Fortvilet', 'Forventningsfull', 'Forventet', 'Fredelig', 'Fredet', 'Frigjørende', 'Frigjort', 'Fristet', 'Fryktet', 'Fryktsom', 'Fungerende', 'Følsom', 'Gledelig', 'Glad', 'Godkjent', 'Grunnleggende', 'Hjelpende', 'Hjelpsom', 'Hoppende', 'Hyllende', 'Hyllet', 'Håpefull', 'Inkludert', 'Inkluderende', 'Innledende', 'Inspirert', 'Inspirerende', 'Interessant', 'Interessert', 'Ivrig', 'Jublende', 'Kommuniserende', 'Konkret', 'Krevende', 'Kritisk ', 'Kritiserende', 'Kritisert', 'Kvalifisert', 'Ledende', 'Lekende', 'Lengtende', 'Lengselsfull', 'Lesende', 'Lett', 'Lidende', 'Livlig', 'Lykkelig', 'Lærerik', 'Lærd', 'Lovende', 'Militær', 'Mistenksom', 'Motivert', 'Motiverende', 'Nasjonal', 'Observant', 'Observerende', 'Opplevelsesrik', 'Opplysende', 'Opplyst', 'Overbevist', 'Overbevisende', 'Praktisk', 'Produktiv', 'Provosert', 'Påståelig', 'Realistisk', 'Realisert', 'Refererende', 'Reflekterende', 'Reflektert', 'Regjerende', 'Regulert', 'Regulerende', 'Repeterende', 'Repetitiv', 'Respektert', 'Samlende', 'Samlet', 'Syngende', 'Sikker', 'Sikret', 'Sjenert', 'Skildrende', 'Skuffet', 'Sliten', 'Springende', 'Spørrende', 'Stigende', 'Stimulerende', 'Strevende', 'Strukturert', 'Studerende', 'Støttende', 'Symbolsk', 'Systematisk', 'Takknemlig', 'Tellende', 'Tankefull', 'Tilgivende', 'Tolerant', 'Trent', 'Trivelig', 'Trygg', 'Tvilende', 'Undersøkende', 'Utdannet', 'Uttalende', 'Utviklet', 'Utøvende', 'Variert', 'Varierende', 'Voksende', 'Økende', 'Økonomisk', 'Ønskelig', 'Ønskende', 'Ønsket', 'Åpenhjertig'];
        //['Snill', 'Kul', 'Høy', 'Lav', 'Redd', 'Modig'];

        var colors = ['Rosa', 'Lyserosa', 'Lillahvit', 'Gammelrosa', 'Halvmørkrosa', 'Rødrosa', 'Rødfiolett', 'Orkide', 'Lysgrålilla', 'Plommefarget', 'Fiolett', 'Magentarød', 'Mørkelilla', 'Purpur', 'Middelsorkide', 'Mørkfiolett', 'Mørkorkide', 'Indigoblå', 'Blåfiolett', 'Middelspurpur', 'Lyskornblå', 'Kornblå', 'Mørkkornblå', 'Lavendel', 'Egghvit', 'Blå', 'Middelsblå', 'Midnattsblå', 'Mørkeblå', 'Marineblå', 'Kongeblå', 'Lyskongeblå', 'Skifergrå', 'Lyshavblå', 'Blåhvit', 'Stålblå', 'Lyshimmelblå', 'Himmelblå', 'Mørkhimmelblå', 'Lyseblå', 'Lysgrønnblå', 'Asurblå', 'Lyscyanblå', 'Lysturkisblå', 'Cyanblå', 'Mørkturkis', 'Middelsturkis', 'Lyshavgrønn', 'Turkis', 'Akvamarin', 'Middelsakvamarin', 'Lysgressgrønn', 'Grønnhvit', 'Vårgrønn', 'Middelshavgrønn', 'Havgrønn', 'Lysegrønn', 'Mørkhavgrønn', 'Gressgrønn', 'Skoggrønn', 'Løvgrønn', 'Gulgrønn', 'Mørkolivengrønn', 'Grønngul', 'Middelsolivengrønn', 'Beige', 'Lysfersken', 'Elfenbenshvit', 'Lysegul', 'Gul', 'Olivengrønn', 'Mørkkaki', 'Fersken', 'Lyskorngul', 'Kaki', 'Gullgul', 'Lyskorngul', 'Gyldenris', 'Mørkebrun', 'Blomsterhvit', 'Kniplingshvit', 'Varmsienna', 'Lysgulbrun', 'Branngul', 'Lyspapaya', 'Gulbrun', 'Gammelhvit', 'Mørkoransje', 'Linhvit', 'Lysnøttebrun', 'Brunrosa', 'Sandbrun', 'Sjokoladebrunt', 'Mørknøttebrun', 'Skjellhvit', 'Nøttebrun', 'Lyslakserød', 'Korallrød', 'Rødoransje', 'Tomatrød', 'Lysrosenrød', 'Lakserød', 'Snøhvit', 'Lyskorallrød', 'Brunlilla', 'Rustrød', 'Rød', 'Brun', 'Mursteinsrød', 'Mørkerød', 'Rødbrun', 'Hvit', 'Røykhvit', 'Lysegrå', 'Sølvgrå', 'Grå', 'Mørkegrå', 'Svart'];

        //['Rød', 'Brun', 'Gul', 'Lilla', 'Kaki', 'Blå', 'Oransje', 'Purpur', 'Eggehvit', 'Rosa', 'Himmelblå'];

        function generateUsername(dictionary = null) {
            if (dictionary == null) {
                return adjectives[Math.floor(Math.random() * adjectives.length)] + colors[Math.floor(Math.random() * colors.length)] + animals[Math.floor(Math.random() * animals.length)]
            } else {
                var username = adjectives[Math.floor(Math.random() * adjectives.length)] + colors[Math.floor(Math.random() * colors.length)] + animals[Math.floor(Math.random() * animals.length)];
                if (dictionary.hasOwnProperty(username)) {
                    return generateUsername(dictionary)
                } else {
                    return username
                }
            }
        }

        db.serialize(function () {
            db.run("BEGIN TRANSACTION");
            db.all("SELECT username FROM Users", function (err, row) {
                if (err === null) {
                    if (row.length === 0) {
                        var returned_username = generateUsername();
                        distributed_usernames[returned_username] = true;
                        res.json({
                            success: true,
                            username: returned_username
                        });
                    } else {
                        for (var i = 0; i < row.length; i++) {
                            distributed_usernames[row[i].username] = true;
                        }
                        var username = generateUsername(distributed_usernames);
                        distributed_usernames[username] = true;
                        res.json({
                            success: true,
                            username: username
                        });
                    }
                } else {
                    logger.log({
                        timestamp: new Date().getTime(),
                        level: 'error',
                        message: 'Error occurred in #cc57',
                        exception: err
                    });
                    res.status(500).json({message: 'Something went wrong.', errorCode: "#cc57"})
                }
            });
            db.run("COMMIT TRANSACTION");
        });
        db.close();
    }
});


// Authenticates the user (login)
app.post('/api/authenticate', function (req, res) {
    if (req.body) {
        var db = new sqlite3.Database('ccdb01.db');
        var stmt = db.prepare('SELECT * FROM Users WHERE username = ?');
        var stmt2 = db.prepare('UPDATE Users SET token = ? WHERE username = ?');
        db.run("BEGIN TRANSACTION");
        db.serialize(function () {
            stmt.all(req.body.username, function (err, row) {
                if (err) {
                    stmt2.finalize();
                    stmt.finalize();
                    db.run("COMMIT TRANSACTION");
                    db.close();
                    logger.log({
                        timestamp: new Date().getTime(),
                        level: 'error',
                        message: 'Error occurred in #cc47',
                        exception: err
                    });
                    res.status(500).json({message: 'Something went wrong.', errorCode: '#cc47'});
                }else {
                    if (row.length == 0) {
                        stmt2.finalize();
                        stmt.finalize();
                        db.run("COMMIT TRANSACTION");
                        db.close();
                        res.status(401).json({message: 'invalid username or password supplied'})
                    }
                    else{
                        bcrypt.compare(req.body.password, row[0].password, function(berr, bres) {
                            if(bres === true) {
                                //password matches, generate token

                                var JWT = signJWT(req.body.username, 10800); //3 hours
                                var token = JWT[0];
                                var expires_at = JWT[1];
                                stmt2.run(token, row[0].username);
                                stmt2.finalize();
                                stmt.finalize();
                                db.run("COMMIT TRANSACTION");
                                db.close();
                                res.json({
                                    success: true,
                                    message: 'user_valid',
                                    authenticated: bres,
                                    token: token,
                                    expires_at: expires_at,
                                    username: row[0].username
                                });
                            }else {
                                // Invalid password
                                stmt2.finalize();
                                stmt.finalize();
                                db.run("COMMIT TRANSACTION");
                                db.close();
                                res.status(401).json({message: 'invalid username or password supplied'})
                            }
                        });
                    }
                }
            });

        });
    }else {
        res.status(400).json({message: 'Failed to authenticate user'});
    }
});

app.post('/api/create-user', function (req, res) {
    if (req.body) {
        var db = new sqlite3.Database('ccdb01.db');
        if (distributed_usernames.hasOwnProperty(req.body.username) && valid_age_groups.hasOwnProperty(req.body.age)
            && valid_eduaction.hasOwnProperty(req.body.education) && valid_genders.hasOwnProperty(req.body.gender)) {
            // use bcrypt to hash password
            var stmt = db.prepare('INSERT INTO Users (username, password, age, gender, education, code_distributed) VALUES (?, ?, ?, ?, ?, ?)');
            bcrypt.hash(req.body.password, 5, function(err, bcryptedPassword) {
                if (err) {
                    //console.log('bcrypt error: ', err);
                    logger.log({
                        timestamp: new Date().getTime(),
                        level: 'error',
                        message: 'Error occurred in #cc51',
                        exception: err
                    });
                    stmt.finalize();
                    res.status(500).json({message: 'Something went wrong', errorCode: '#cc51'})
                }else {
                    db.serialize(function () {
                        stmt.run(req.body.username, bcryptedPassword, req.body.age, req.body.gender, req.body.education, 0, function (error) {
                            if (error) {
                                logger.log({
                                    timestamp: new Date().getTime(),
                                    level: 'error',
                                    message: 'Error occurred in #cc03',
                                    exception: error
                                });
                                res.status(500).json({message: 'Something went wrong.', errorCode: '#cc03'})
                            }
                            else {
                                res.json({
                                    success: true,
                                    message: 'user created'
                                })
                            }
                        });
                        stmt.finalize();
                    });
                    db.close();
                }
            });
        }else {
            logger.log({
                timestamp: new Date().getTime(),
                level: 'info',
                message: 'create-user tried with invalid info (2nd else)',
                values: [req.body.username, req.body.age, req.body.education, req.body.gender]
            });
            res.status(400).json({message: 'Invalid information supplied'})
        }
    }else {
        logger.log({
            timestamp: new Date().getTime(),
            level: 'info',
            message: 'create-user tried with invalid info (3rd else)'
        });
        res.status(400).json({message: 'failed to create user'});
    }
});

app.get('/api/current-user', jwtCheck, function (req, res) {
    //console.log('req auth token: ', req.headers['authorization']);
    var token = parseToken(req.headers['authorization']);
    if (token == null) {
        res.status(401).json({message: 'invalid token'})
    }else {
        var db = new sqlite3.Database('ccdb01.db');
        var stmt = db.prepare('SELECT username FROM Users WHERE token = ?');
        db.serialize(function () {
            stmt.all(token, function (err, row) {
                if (err) {
                    logger.log({
                        timestamp: new Date().getTime(),
                        level: 'error',
                        message: 'Error occurred in #cc04',
                        exception: err
                    });
                    res.status(500).json({message: 'Something went wrong', errorCode: '#cc04'})
                }
                else if (row.length === 0) {
                    logger.log({
                        timestamp: new Date().getTime(),
                        level: 'info',
                        message: 'current-user not found',
                        token: token
                    });
                    res.status(400).json({message: 'No user found.'})
                }
                else {
                    res.json({
                        success: true,
                        username: row[0].username
                    })
                }
            });
            stmt.finalize();
        });
        db.close();
    }
});

app.post('/api/logout', jwtCheck, function (req, res) {
    var token = parseToken(req.headers['authorization']);
    getInternalUsername(req.headers['authorization'], function (username) {
        if(!username) {
            logger.log({
                timestamp: new Date().getTime(),
                level: 'info',
                message: 'logout: tried to logout without username',
                token: token
            });
            res.status(403).json({message: 'invalid token'})
        }
        else {
            var db = new sqlite3.Database('ccdb01.db');
            var stmt = db.prepare('UPDATE Users SET token = null WHERE username = ? AND token = ?');
            db.serialize(function () {
                db.run("BEGIN TRANSACTION");
                stmt.run(username, token);
                stmt.finalize();
                db.run("COMMIT TRANSACTION");
            });
            db.close();
            res.status(200).json({message: 'logged_out'})
        }
    });


});


app.get('/api/questions/:numq', jwtCheck, function(req, res){
    var num = req.params.numq;
    //console.log('API received numq=' + num);
    getInternalUsername(req.headers['authorization'], function (username) {
        //console.log('got username in numq: ', username);
        var message = '';
        if (!username) {
            message = 'User is ' + username;
            logger.log({
                timestamp: new Date().getTime(),
                level: 'info',
                message: '403 error in /questions/:numq',
                exception: message
            });
            res.status(403).json({message: 'invalid token'});
        }
        else if (isNaN(num)) {
            message = 'Invalid parameter number';
            logger.log({
                timestamp: new Date().getTime(),
                level: 'info',
                message: '400 isNaN in /questions/:numq',
                exception: message
            });
            res.status(400).json({message: message})
        }
        else {
            // Round: 5-9: 1, 10-14: 2, 15-19: 3 osv
            // Ceil: 5: 1, 6-10: 2, 11-15: 3, 16-20: 4 osv
            // Ser ut som round gir litt for få
            //console.log('Checks are ok, starting db queries...');
            var num_baseline_questions = Math.ceil(num / 5);
            var new_num = num - num_baseline_questions;
            //console.log("Num: " + num + ", numBaseline: " + num_baseline_questions + ", Newnum: " + new_num);
            var db = new sqlite3.Database('ccdb01.db');

            const stmt = db.prepare("SELECT id, claim FROM ControlClaims WHERE id NOT IN (SELECT control_id from Answers WHERE control_id IS NOT NULL AND username = ?)");
            // Returns the most answered questions until a limit (answer_count_limit)
            const stmt2 = db.prepare(
                "SELECT uc1.id, uc2.claim FROM (" +
                    "SELECT uc3.id FROM (" +
                        "SELECT uc4.id, COUNT(uc4.id)-1 as answerCount FROM UnlabeledClaims as uc4 GROUP BY uc4.id " +
                        "UNION " +
                        "SELECT a1.unlabeled_id, COUNT(a1.unlabeled_id) as answerCount FROM Answers as a1 WHERE a1.unlabeled_id IS NOT NULL AND a1.answer != 3 GROUP BY a1.unlabeled_id" +
                    ") AS uc3 " +
                    "WHERE uc3.id NOT IN (SELECT a2.unlabeled_id FROM Answers as a2 WHERE a2.unlabeled_id IS NOT NULL AND a2.username = ?) " +
                    "GROUP BY uc3.id HAVING uc3.answerCount <= ? ORDER BY uc3.answerCount DESC LIMIT ?" +
                ") AS uc1 "+
                "INNER JOIN UnlabeledClaims as uc2 ON uc1.id=uc2.id ");

            //Get the ? least answered questions NOT already answered by the current user
            const stmt3 = db.prepare(
                "SELECT uc1.id, uc2.claim FROM (" +
                    "SELECT uc3.id FROM (" +
                        "SELECT uc4.id, COUNT(uc4.id)-1 as answerCount FROM UnlabeledClaims as uc4 GROUP BY uc4.id " +
                        "UNION " +
                        "SELECT a1.unlabeled_id, COUNT(a1.unlabeled_id) as answerCount FROM Answers as a1 WHERE a1.unlabeled_id IS NOT NULL AND a1.answer != 3 GROUP BY a1.unlabeled_id" +
                    ") AS uc3 " +
                    "WHERE uc3.id NOT IN (SELECT a2.unlabeled_id FROM Answers as a2 WHERE a2.unlabeled_id IS NOT NULL AND a2.username = ?) " +
                    "GROUP BY uc3.id ORDER BY uc3.answerCount ASC LIMIT ?" +
                ") AS uc1 "+
                "INNER JOIN UnlabeledClaims as uc2 ON uc1.id=uc2.id ");

            db.serialize(function () {
                db.run("BEGIN TRANSACTION");

                /*
                 First query: Get all BaselineQuestions NOT already answered by the current user
                 */
                stmt.all(username, function (err, row) {
                    if (err === null) {
                        //console.log('baseline row: ', row);
                        var baseline = row;
                        var questions = null;
                        if (row.length >= num_baseline_questions) {
                            /*
                             Have enough baseline questions, run statement 2, with LIMIT new_num
                             Second query: Get the most answered questions until a limit (answer_count_limit). If this do not
                             give correct amount of questions run third query
                             Third query: Get the N(new_num) least answered questions NOT already answered by the current user
                             */
                            stmt2.all(username, answer_count_limit, new_num, function (err, row1) {
                                if (err === null) {
                                    if(row1.length < new_num){
                                        stmt3.all(username, new_num, function (err, row2) {
                                            if (err === null) {
                                                //console.log("Questions: ", row2);
                                                questions = row2;

                                                var json = shuffleQuestionsWithBaseline(questions, baseline, num_baseline_questions);
                                                //console.log('json: ', json);
                                                res.json(json);
                                            }
                                            else {
                                                logger.log({
                                                    timestamp: new Date().getTime(),
                                                    level: 'error',
                                                    message: 'Error occurred in #cc11',
                                                    exception: err
                                                });
                                                res.status(500).json({message: "Something went wrong", errorCode: "#cc11"});
                                            }
                                        });
                                        stmt3.finalize();
                                    }
                                    else {
                                        //console.log("Questions: ", row1);
                                        questions = row1;
                                        stmt3.finalize();
                                        var json = shuffleQuestionsWithBaseline(questions, baseline, num_baseline_questions);
                                        //console.log('json: ', json);
                                        res.json(json);
                                    }
                                }else {
                                    logger.log({
                                        timestamp: new Date().getTime(),
                                        level: 'error',
                                        message: 'Error occurred in #cc98',
                                        exception: err
                                    });
                                    res.status(500).json({message: "Something went wrong", errorCode: "#cc98"});
                                }
                            });
                        } else {
                            /*
                             Not enough baseline questions, run statement 2, with LIMIT num - row.length
                             Second query: Get the most answered questions until a limit (answer_count_limit). If this do not
                             give correct amount of questions run third query
                             Third query: Get the N(num - row.length) least answered questions NOT already answered by the current user
                             */
                            stmt2.all(username, answer_count_limit, (num - row.length), function (err, row1) {
                                if (err === null) {
                                    //console.log("Row1: ", row1);
                                    if(row1.length < new_num){
                                        stmt3.all(username, new_num, function (err, row2) {
                                            if (err === null) {
                                                //console.log("Questions: ", row2);
                                                questions = row2;

                                                var json = shuffleQuestionsWithBaseline(questions, baseline, num_baseline_questions);
                                                //console.log('json: ', json);
                                                res.json(json);
                                            }
                                            else {
                                                logger.log({
                                                    timestamp: new Date().getTime(),
                                                    level: 'error',
                                                    message: 'Error occurred in #cc321',
                                                    exception: err
                                                });
                                                res.status(500).json({message: "Something went wrong", errorCode: "#cc321"});
                                            }
                                        });
                                        stmt3.finalize();
                                    }
                                    else {
                                        //console.log("Questions: ", row1);
                                        questions = row1;
                                        stmt3.finalize();
                                        var json = shuffleQuestionsWithBaseline(questions, baseline, num_baseline_questions);
                                        //console.log('json: ', json);
                                        res.json(json);
                                    }
                                }else {
                                    logger.log({
                                        timestamp: new Date().getTime(),
                                        level: 'error',
                                        message: 'Error occurred in #cc68',
                                        exception: err
                                    });
                                    res.status(500).json({message: "Something went wrong", errorCode: "#cc68"});
                                }
                            });
                        }
                        stmt2.finalize();
                    } else {
                        stmt3.finalize();
                        stmt2.finalize();
                        logger.log({
                            timestamp: new Date().getTime(),
                            level: 'error',
                            message: 'Error occurred in /question/:numq (stmt1)',
                            exception: err
                        });
                    }
                });
                stmt.finalize();
                db.run("COMMIT TRANSACTION");
            });
            db.close();
        }
    });

});

app.post('/api/questions/done/', jwtCheck, function (req, res) {
    getInternalUsername(req.headers['authorization'], function (username) {
        var questions = req.body.questions;
        var baseline = req.body.baseline;

        var db = new sqlite3.Database('ccdb01.db');

        /*
         Question data
         */
        //console.log('Inserting answer data...');
        let message = '';
        if (!username) {
            message = 'User is '+ username;
            logger.log({
                timestamp: new Date().getTime(),
                level: 'info',
                message: 'Error occurred in /question/done )',
                exception: message
            });
            res.status(403).json({message: 'invalid token'});
        }
        db.serialize(function () {
            if(questions) {
                var num_q = 0;
                var stmt = db.prepare('INSERT INTO Answers(answer, unlabeled_id, username) VALUES (?, ?, ?)');
                for(var v of questions) {
                    if (!v.answer) {
                        message = 'Answer is ' + v.answer;
                        logger.log({
                            timestamp: new Date().getTime(),
                            level: 'error',
                            message: 'Error occurred in /question/done (questions)',
                            exception: message
                        });
                        res.status(400).json({message: message});
                    }else {
                        try {
                            stmt.run(v.answer, v.id, username);
                            num_q +=1;
                        }catch (err) {
                            logger.log({
                                timestamp: new Date().getTime(),
                                level: 'error',
                                message: 'Error occurred in #cc91',
                                exception: err
                            });
                            res.status(500).json({message: "Something went wrong", errorCode: "#cc91"});
                        }
                    }
                }
                stmt.finalize();
                //console.log(`Done inserting ${num_q} answer rows for questions.\n`);

            }

            /*
             Baseline data
             */
            if(baseline) {
                var num_q = 0;
                var stmt2 = db.prepare('INSERT INTO Answers(answer, control_id, username) VALUES (?, ?, ?)');
                for(var b of baseline) {
                    if (!b.answer) {
                        message =  'Answer is ' + b.answer;
                        logger.log({
                            timestamp: new Date().getTime(),
                            level: 'error',
                            message: 'Error occurred in /question/done (baseline)',
                            exception: message
                        });
                        res.status(400).json({message: message});
                    }else {
                        try {
                            stmt2.run(b.answer, b.id, username);
                            num_q +=1;
                        }catch (err) {
                            logger.log({
                                timestamp: new Date().getTime(),
                                level: 'error',
                                message: 'Error occurred in #cc123',
                                exception: err
                            });
                            res.status(500).json({message: "Something went wrong", errorCode: "#cc123"});
                        }
                    }
                }
                stmt2.finalize(() => {
                    if (username) {
                        //console.log('Calculating user score...');
                        calculateUserScore(username);
                    }else {
                        logger.log({
                            timestamp: new Date().getTime(),
                            level: 'info',
                            message: 'Error occurred in /question/done (stmt2 finalize)',
                            exception: 'Error: No user defined'
                        });
                    }
                });
                //console.log(`Done inserting ${num_q} answer rows for baseline.`);
            }

        });
        db.close();
        res.json({message: 'API: Items added'});

    });

});

app.get('/api/scores/', jwtCheck, function (req, res) {
    //console.log('In get scores');
    getInternalUsername(req.headers['authorization'], function (username) {
        if (!username) {
            var message = 'User is ' + username;
            logger.log({
                timestamp: new Date().getTime(),
                level: 'info',
                message: '403 Error occurred in /scores',
                exception: message
            });
            res.status(403).json({message: 'invalid token'});
        }
        else{
            var db = new sqlite3.Database('ccdb01.db');

            db.serialize(function () {
                db.run("BEGIN TRANSACTION");
                db.all("SELECT u.username, u.score, u.code_distributed, count(a.unlabeled_id) as q_count, count(a.control_id) as b_count FROM Users u" +
                    " INNER JOIN Answers as a ON u.username=a.username WHERE a.answer != 3 GROUP BY u.username", function (err, row) {
                    if (err === null) {

                        // Calculate "frontend" scores for each user
                        row.map(obj => {
                            obj.answer_count = obj.q_count + obj.b_count;
                            obj.score = `${Math.round(obj.score*100)}%`; //score in percentage

                        });

                        //console.log(row);
                        //console.log('UserScore row: ', row);
                        res.json({'scores': row});
                    } else {
                        logger.log({
                            timestamp: new Date().getTime(),
                            level: 'error',
                            message: 'Error occurred in #cc61',
                            exception: message
                        });
                        res.status(500).json({message: "Something went wrong", errorCode: "#cc61"});
                    }
                });
                db.run("COMMIT TRANSACTION");
            });
            db.close();
        }
    });

});

app.get('/api/generate-code/', jwtCheck, function (req, res) {
    //console.log('In get scores');
    getInternalUsername(req.headers['authorization'], function (username) {
        if (!username) {
            var message = 'User is ' + username;
            logger.log({
                timestamp: new Date().getTime(),
                level: 'info',
                message: '403 Error occurred in /generate-code',
                exception: message
            });
            res.status(403).json({message: 'invalid token'});
        }
        else{
            function generateDistCode(dictionary=null) {
                function s4() {
                    return Math.floor((1 + Math.random()) * 0x10000)
                        .toString(16)
                        .substring(1);
                }
                var c = s4() + s4() + '-' + s4() + '-' + s4() + '-' + s4() + '-' + s4() + s4() + s4();

                if (dictionary == null) {
                    return c
                }
                else {
                    if(dictionary.hasOwnProperty(c)){
                        return generateDistCode(dictionary)
                    }
                    else {
                        return c
                    }
                }
            }

            var db = new sqlite3.Database('ccdb01.db');
            var stmt = db.prepare("SELECT code_distributed FROM Users WHERE username = ?");

            db.serialize(function () {
                db.run("BEGIN TRANSACTION");
                stmt.all(username, function (err, row) {
                    if (err === null) {
                        if(row.length === 1){
                            if(row[0].code_distributed === 0){
                                var code = null;
                                db.all("SELECT code FROM Distcodes", function (err, row) {
                                    if(err === null){
                                        if (row.length === 0) {
                                            var returned_code = generateDistCode();
                                            distributed_distcodes[returned_code] = true;
                                            code = returned_code;
                                        } else {
                                            for (var i = 0; i < row.length; i++) {
                                                distributed_distcodes[row[i].code] = true;
                                            }
                                            var returned_code = generateDistCode(distributed_distcodes);
                                            distributed_distcodes[returned_code] = true;
                                            code = returned_code;
                                        }
                                        var stmt2 = db.prepare("INSERT INTO Distcodes VALUES (?)");
                                        stmt2.run(code);
                                        stmt2.finalize();

                                        var stmt3 = db.prepare("UPDATE Users SET code_distributed = ? WHERE username = ?");
                                        stmt3.run(1, username);
                                        stmt3.finalize();

                                        res.json({'distcode': code})
                                    } else {
                                        logger.log({
                                            timestamp: new Date().getTime(),
                                            level: 'error',
                                            message: 'Error occurred in /generate-code (stmt)',
                                            exception: message
                                        });
                                        //console.error(err);
                                    }
                                });

                            }
                            else {
                                //res.json({'distcode': 'no'});
                                res.json({message: 'Code already distributed.'})
                            }
                        }
                        else {
                            logger.log({
                                timestamp: new Date().getTime(),
                                level: 'error',
                                message: 'Error occurred in #cc65',
                            });
                            res.status(500).json({message: 'Something went wrong.', errorCode: '#cc65'});
                        }
                    } else {
                        logger.log({
                            timestamp: new Date().getTime(),
                            level: 'error',
                            message: 'Error occurred in #cc23',
                            exception: err
                        });
                        res.status(500).json({message: 'Something went wrong.', errorCode: '#cc23'});
                    }
                });
                stmt.finalize();
                db.run("COMMIT TRANSACTION");
            });
            db.close();
        }
    });

});




app.post('/api/logger/', jwtCheck, function (req, res) {
    if (req.body && req.body.data) {
        // Parse logs
        var logs = JSON.parse(req.body.data);
        Object.keys(logs).forEach(function (key) {
            frontend_logger.log({
                timestamp: logs[key].timestamp,
                level: logs[key].level.toLowerCase(),
                message: logs[key].message,
                exception: logs[key].exception,
                url: logs[key].url
            });
        });
        res.json({message: 'logged'});
    }else {
        logger.log({
            timestamp: new Date().getTime(),
            level: 'error',
            message: 'Error occurred in #cc02',
        });
        res.status(500).json({message: 'Something went wrong.', errorCode: '#cc02'});
    }

});

console.log(`Listening on port ${port}...`);
app.listen(port);